local sysinfo = require "api_sysinfo"
local MSG = require "api_msg"
local CO = require "api_coservice"
local cjson = require "cjson"

local cached_versions
local cached_sn

local dset = {}

local function API_server_info( args)
	local cpu_c = sysinfo.GetCpuCores()
	local cpu_p, cpu_i = sysinfo.GetCpuLoad()
	local mem_f, mem_t = sysinfo.GetMemUse()
	local pers_m, pers_t = sysinfo.GetPersisTime()

	local info = {
		addr = ngx.var.server_addr,
		name = ngx.var.scheme .. "://" .. ngx.var.host,
		port = ngx.var.server_port,
		persis = "" .. math.floor(pers_m/3600) .. "H " .. math.floor((pers_m%3600)/60) .. "M " .. math.floor(pers_m%60) .. "S",
		start_time = string.format( "%04d-%02d-%02d %02d:%02d:%02d", pers_t.year, pers_t.month, pers_t.day, pers_t.hour, pers_t.min,  pers_t.sec ),
		cpu_cores = cpu_c,
		cpu_payload = math.floor(cpu_p*100)/100,
		mem_used = math.floor( ((mem_t - mem_f)/1024)*100 )/100,
		mem_total = math.floor( (mem_t/1024)*100 ) / 100
	}

	MSG.OK( "", { data = info } )
	return true
end

local function API_reset(args)
	local proxy = CO.GetProxy( "/", "systemctrl" )
	if not proxy then
		MSG.ERROR( "systemError" )
		return true
	else
		proxy.Reset()
		proxy:Destroy()
		MSG.OK()
		return true
	end
end

local function API_reboot(args)
	local proxy = CO.GetProxy( "/", "systemctrl" )
	if not proxy then
		MSG.ERROR( "systemError" )
	else
		-- local dev = require "api_device"
		-- dev.REBOOTING = true
		proxy.Reboot(2)
		proxy:Destroy()
		MSG.OK()
		return true
	end
	return true
end

local function API_restore(args)
	local proxy = CO.GetProxy( "/", "systemctrl" )
	if not proxy then
		MSG.ERROR( "systemError" )
	else
		-- local dev = require "api_device"
		-- dev.REBOOTING = true
		proxy.RestoreFactory()
		proxy:Destroy()
		MSG.OK()
	end
	return true
end

local function API_reconnect()
	-- local proxy = CO.GetProxy("/", "codec")
	-- if not proxy then
	-- 	MSG.ERROR("codecError")
	-- 	return true
	-- end

	-- local is_ok = proxy.Reset()
	-- proxy:Destroy()
	-- if not is_ok then
	-- 	MSG.ERROR("connectNdiError")
	-- 	return true
	-- end
	
	-- MSG.OK("")
    MSG.ERROR("Not have this function")
	return true
end

local function API_status(args)
	if not cached_versions or not cached_sn then
		local proxy = CO.GetProxy( "/", "systemctrl" )
		if proxy then
			local r
			r, cached_versions = proxy.GetVersion()
			r, cached_sn = proxy.GetSerialNo()
			proxy:Destroy()
		end
	end

	local cpu_c = sysinfo.GetCpuCores()
	local cpu_p, cpu_i = sysinfo.GetCpuLoad()
	local mem_f, mem_t = sysinfo.GetMemUse()
	local pers_m, pers_t = sysinfo.GetPersisTime()
	-- software_version  hardware_version firmware_version
	local info = {
		softwareVersion = cached_versions and cached_versions.SOFTWARE_VERSION or "1.00.0000",
		hardwareVersion = cached_versions and cached_versions.HARDWARE_VERSION or "1.0",
        firmwareVersion  = cached_versions and cached_versions.FIRMWARE_VERSION or "0.00",

		mem_total = math.floor((mem_t/1024)*100)/100,
		mem_free = math.floor((mem_f/1024)*100)/100,
		cpu_payload = math.floor(cpu_p*100)/100,
		persis_time = pers_m,
		serial_no = cached_sn or "00000000",
		cpu_cores = cpu_c
	}

	MSG.OK( "", { data = {info} } )
	return true
end

local function API_newsn(args)
	if not args or not args.serial_no then
		MSG.ERROR( "argumentError" )
	else
		local proxy = CO.GetProxy( "/", "systemctrl" )
		if proxy then
			cached_sn = nil
			local r, done = proxy.CreateSerialNo(args.serial_no)
			proxy:Destroy()
			if r and done then
				MSG.OK( "" )
			else
				MSG.ERROR( "createSerialNoError" )
			end
		else
			MSG.ERROR( "systemError" )
		end
	end
	return true
end

local function API_GetHostname(args)
	local proxy = CO.GetProxy( "/webDaemon", "systemctrl" )
	if not proxy then
		MSG.ERROR("webDaemon proxy Error!")
		return
	end
	local is_ok, hostname = proxy.GetHostname()
	proxy:Destroy()
	if not is_ok then
		MSG.ERROR("Get Hostname Error")
		return
	end
	MSG.OK("", {hostname = hostname})
	return true
end

local function API_SetHostname(args)
	if not tostring(args.hostname) or tostring(args.hostname) == "" then
		MSG.ERROR("Invalid Hostname")
		return
	end
	local proxy = CO.GetProxy( "/webDaemon", "systemctrl" )
	if not proxy then
		MSG.ERROR("webDaemon proxy Error!")
		return
	end
	local is_ok, name = proxy.SetHostname(tostring(args.hostname))
	proxy:Destroy()
	if not is_ok or not name then
		MSG.ERROR("Set Hostname Error")
		return
	end
	MSG.OK("")
	return true
end

local function API_Ping(args)
	local addr = tostring(args.targetAddr)
	if not addr or addr == "" then
		MSG.ERROR("Invalid addr!")
		return 
	end
	local proxy = CO.GetProxy( "/", "systemctrl" )
	if not proxy then
		MSG.ERROR("systemCtrl proxy Error!")
		return
	end
	local is_ok, ok, reach, loss = proxy.Ping(addr)
	proxy:Destroy()
	if not is_ok then
		MSG.ERROR("systemError")
	elseif not ok or not loss then
		MSG.OK("", {data = {reachable = false, loss_rate = 100}})
	elseif reach and loss < 100 then
		MSG.OK("", {data = {reachable = true, loss_rate = loss}})
	else
		MSG.OK("", {data = {reachable = false, loss_rate = loss}})
	end
	return true
end

-----------------------------------------------------------

dset["server_info"] = API_server_info
dset["reset"] = API_reset
dset["reboot"] = API_reboot
dset["restore"] = API_restore
dset["reconnect"] = API_reconnect
dset["status"] = API_status
dset["newsn"] = API_newsn
dset["GetHostname"] = API_GetHostname
dset["SetHostname"] = API_SetHostname
dset["Ping"] = API_Ping
return dset
