#!/usr/bin/lua
local coaf = require "coaf.core"
local _dbg = require "coaf.debug"
local service = require "coaf.service"
local thread = require "coaf.thread"
local event = require "coaf.event"
local mod_codec, mod_Streamer, mod_OSD, mod_audioEnc, mod_Webrtc, mod_record

coaf.ChDir(coaf.GetAppDir(arg))

local withLog = false
for i = 1, #arg do
    if arg[i] == "--log" then
        withLog = true
        break
    end
end

if withLog then
    local roll = require 'coaf.log.rollfile'
    os.execute('/bin/mkdir -p /var/log/')
    local logF = roll.Create('/var/log/codec.log')
    if not logF or not logF:init() then
        _dbg.Error('ENCODER', 'Fail to open log file: /var/log/encoder.log, so the log messages should be output to console.')
        withLog = false
    else
        _dbg.AddLogger('encoderlogfile', logF, nil, nil)
    end
end

coaf.Init("codec")

if withLog then
	local stdio = require "coaf.log.stdio"
	stdio.SetDefaultFilterName( "ENCODER" )
	stdio.Start()
end

_G.streamer_event = event.createNew( "streamer", true )

local type_change = {
    number = tonumber,
    string = tostring
}

local function empty_function() print("empty function") return true end
local function errror_function() print("error function") return false end
local function createResult(...)
    local result = {...}
    local is_ok = result[1]
    if not is_ok then
        print("###########XPCALL-ERROR#######", result[2])
        return
    end
    return table.unpack(result, 2, #result)
end

local function createNewObj(source, name)
    if not source or type(source) ~= "table" then
        return
    end
    
    return setmetatable({}, {
        __index = function(_, key)
            local func = source[key]
            if not func then
                print("not have this func and return", name)
                return empty_function
            end

            return function(...)
                local arguments, is_dict
                local tbl = {...}
                if type(tbl[1]) == "table" then
                    is_dict = true
                    arguments = tbl[1]
                else
                    arguments = tbl
                end
            
                local rule = source.Rule[key]
                -- print("key(function name) = ", key, name)
                if not rule then
                    print("not this rule and return empty function", name)
                    return empty_function()
                end
            
                if not rule.is_open then
                    --NOTE: 这里应该区别是网络请求还是内部coaf调用，如果是网络请求就return
                    print("not open this function: ", key, name)
                    return empty_function()
                end
            
                local args_list = {}
                for i, v in ipairs(rule.parameters) do
                    local value
                    if arguments[v.key] ~= nil then
                        value = arguments[v.key]
                    else
                        value = arguments[i]
                    end
                    if v.must and not value then
                        print("not have the parameters : ", v.key, name)
                        return errror_function()
                    end
            
                    if value ~= nil then
                        local change_func = type_change[v.type]
                        if change_func then
                            value = change_func(value)
                        end
                        if type(value) ~= v.type  then
                            print("parameters type is not allowed : ", v.key, name)
                            return errror_function()
                        end
                    end
            
                    args_list[i] = value
                end
            
                if is_dict then
                    return createResult(xpcall(function()
                        return source[key](table.unpack(args_list, 1, #rule.parameters))
                    end, debug.traceback))
                else
                    return createResult(xpcall(function()
                        return source[key](table.unpack(tbl, 1, #tbl))
                    end, debug.traceback))
                end
            end
        end,
    })
end

function table.merge(old, new)
    if type(old) ~= "table" or type(new) ~= "table" then
        return
    end
    for i, v in pairs(new) do
        old[i] = v
    end
end

function table.update(old, new)
    for i, v in pairs(old) do
        if type(new[i]) == "table" and type(v) == "table" then
            table.update(v, new[i])
        elseif type(v) ~= "table" and new[i] ~= nil then
            old[i] = new[i]
        end
    end
end

function table.copy(src)
    if type(src) ~= "table" then
        return
    end

    local new = {}
    for k, v in pairs(src) do
        if type(v) == "table" then
            new[k] = table.copy(v)
        else
            new[k] = v
        end
    end
    return new
end

function table.compare(a, b)
    for i, v in pairs(a) do
        if type(v) ~= "table" then
            if b[i] ~= nil and b[i] ~= v then
                print("v = ", v, " i = ", i)
                return true
            end
        else
            if b[i] and table.compare(v, b[i]) then
                return true
            end
        end
    end
end

_G.codec_cfg = require("configs.codec_config").Load()
_G.osd_cfg = require("configs.osd_config").Load()
_G.streamer_cfg = require("configs.streamer_config").Load()
_G.capture_cfg = require ("configs.capture_config").Load()
_G.aenc_cfg = require ("configs.aenc_config").Load()
_G.vpss_cfg = require ("configs.vpss_config").Load()
_G.media_users = require ("configs.medusers_config").Load()

service.InitSocket()

local for500ms_Timer, for5s_Timer, mod_OSD_adapter, for1s_tTimer
local for500ms_timer_working = false
local function for500ms_PeriodCheck()
    if for500ms_timer_working then
        return
    end
    for500ms_timer_working = true

    mod_OSD_adapter = mod_OSD_adapter or require "mod_OSD_adapter"
    mod_OSD_adapter.PeriodRender()

    for500ms_timer_working = false
end

local function for1s_threadTimer()
    mod_record.Check()
end


local restaring = false
_G.SHOULD_REINIT_STEAMER = {}
local function for5s_PeriodCheck()
    if next(_G.SHOULD_REINIT_STEAMER) and not restaring then
        restaring = true
        if mod_Streamer then
            -- mod_Streamer.Destroy()
            for _, v in pairs(_G.SHOULD_REINIT_STEAMER) do
                mod_Streamer.Init(v)
            end
        end
        restaring = false
        _G.SHOULD_REINIT_STEAMER = {}
    end
end

local capture, global, ain, vpss
local function Main(args)
    capture = require "soc.capture"
    vpss = require "soc.vpss"
    global = require "soc.global"
    ain = require "soc.ain"

    mod_codec = require "codec"
    mod_OSD = require "mod_OSD"
    mod_Streamer = require "mod_Streamer"
    mod_audioEnc = require "mod_audioEnc"
    mod_Webrtc = require "mod_Webrtc"
    mod_record = require "mod_record"

    global.Init()
    capture.Init()
    vpss.Init()
    ain.Init(true)
    mod_audioEnc.Init()
    mod_codec.Init()
    mod_OSD.Init()
    mod_OSD.ConstructOSDs()

    mod_Streamer.Init()
    mod_Webrtc.Init()
    mod_record.Init()

    for5s_Timer = coaf.NewTimer(for5s_PeriodCheck, 5, 5)
    if for5s_Timer then
        for5s_Timer:start()
    end

    for500ms_Timer = coaf.NewTimer(for500ms_PeriodCheck, 5, 0.5)
    if for500ms_Timer then
        for500ms_Timer:start()
    end

    for1s_tTimer = thread.NewTimer(for1s_threadTimer, 1, 1)
    if for1s_tTimer then
        for1s_tTimer:start()
    end

    local coservice = service.New("codec", "unix:/var/run/codec.sock")
    if coservice then
        coservice:CreateCoServerProxy()
        coservice:ExposeObject("/codec", createNewObj(mod_codec, "mod_codec"))
        coservice:ExposeObject("/osd", createNewObj(mod_OSD))
        coservice:ExposeObject("/streamer", createNewObj(mod_Streamer, "mod_Streamer"))
        coservice:ExposeObject("/capture", createNewObj(capture, "capture"))
        coservice:ExposeObject("/audio", createNewObj(mod_audioEnc, "mod_audioEnc"))
        coservice:ExposeObject("/mod_Webrtc", createNewObj(mod_Webrtc, "mod_Webrtc"))
        coservice:ExposeObject("/record", mod_record)
    else
        print("~~~~~~~~~Can not create service~~~~~~~~~~")
    end
end

coaf.AddDeadSignal("SEGV", 1)
coaf.AddDeadSignal("HUP", 1)
coaf.Start(Main, args)

if for5s_Timer then
    for5s_Timer:stop()
end

if for500ms_Timer then
    for500ms_Timer:stop()
end

if for1s_tTimer then
    for1s_tTimer:stop()
end

event.free( "codec" )
mod_Streamer.Destroy()
mod_codec.Exit()
mod_audioEnc.Destroy()
vpss.Destroy()
capture.Destroy()
