#!/usr/bin/env lua
local coaf = require "coaf.core"
local _dbg = require "coaf.debug"
local thread = require "coaf.thread"
local service = require "coaf.service"

local socket = require "cosocket.core"
local roll = require "coaf.log.rollfile"

local event = require "coaf.event"

-- Change current dir to app location
coaf.ChDir(coaf.GetAppDir(arg))

------------------------------------------------
local cfg = require("nas_config").Load()

local withLog = false
local _x=1
while _x <= #arg do
	if arg[_x] == "--log" then
		withLog = true
	end
	_x = _x+1
end

-- First, initialize the log file.
if withLog then
	os.execute( "/bin/mkdir -p /var/log" )
	local logF = roll.Create("/var/log/nas_manager.log")
	if not logF or not logF:init() then
		_dbg.Error("NAS", "Fail to open log file: /var/log/nas_manager.log, so the log messages should be output to console.")
	else
		_dbg.AddLogger("logfile", logF, nil, nil)
	end
end


------------------------------------------------
_G.g_service = nil
------------------------------------------------

local nas = {
	mounts = {}
}

local function mountThread(thrObj, id, ctx, item)
	--TODO: Check NAS type
	local nas_mounter = require "nas_NFS"
	local mnt_type = "nfs"

	if item.type and item.type ~= "" and item.type ~= "nfs" then
		if item.type == "cifs" then
			nas_mounter = require "nas_CIFS"
			mnt_type = "cifs"
		end
	end

	_dbg.Info( "NAS", "Start mounting NAS/" .. mnt_type .. " item of ", id , "..." )
	while not thrObj.terminated do
		local mnt,name = nas_mounter.Mount( _G.g_event, id, item )
		if mnt then
			ctx.status = "mounted"
			ctx.mount_thread = nil
			ctx.name = name
			ctx.mount = mnt
			break
		else
			_dbg.Error( "NAS", "Mount NAS/" .. mnt_type .. " item of ", id , " fail! with message:", name )
		end
		coaf.Sleep(3)	--Sleep 3 seconds and try again.
	end

	if ctx.status == "mounted" then
		_dbg.Info( "NAS", "NAS/" .. mnt_type .. " item of ", id , " is mounted at ", ctx.mount, ", name is:", ctx.name )
		os.execute( "/usr/bin/disk-attach.lua " .. ctx.mount .. " nas " .. " " .. ctx.name .. " " .. mnt_type )
	else
		_dbg.Info( "NAS", "NAS/" .. mnt_type .. " item of ", id , " is not mounted!" )
	end
end

function nas.StartMount(id)
	local cfgItem = cfg[id]
	if not cfgItem then
		return false, "Invalid NAS item id"
	else
		local ctx = { status = "mounting" }
		ctx.mount_thread = thread.New( mountThread )
		ctx.type = cfgItem.type
		nas.mounts[id] = ctx
		ctx.mount_thread:Start( id, ctx, cfgItem )
		return true
	end
end

function nas.GetList()
	local rslt = {}
	local k,v 
	for k,v in pairs(cfg) do
		local item = { id = k, name = v.name, type=v.type, host = v.host, volumn = v.volumn, options = v.options, choose = v.choose }
		local ctx = nas.mounts[k]
		if ctx then
			item.status = ctx.status
			item.name = ctx.name or item.name
			item.mount = ctx.mount
		else
			item.status = "unmount"
		end
		table.insert(rslt,item)
	end
	table.sort(rslt, function(a,b)
		return a.id < b.id
	end)
	return rslt
end

function nas.Add( id, name, type, host, volumn, options )
	if not id or not type or not host or not volumn then
		return false, "Invalid NAS arguments"
	end

	if cfg[id] then
		return false, "The NAS item of " .. (id or "?") .." is already exist!"
	end

	cfg[id] = {
		name = name ~= "" and name or id,
		type = type ~= "" and type or "nfs",
		host = host,
		volumn = volumn,
		options = options
	}

	cfg:__TRIGGER(true)
	nas.StartMount(id)
	return true
end

function nas.Remove( id )
	local ctx = nas.mounts[id]
	if ctx then
		if ctx.mount_thread then
			ctx.mount_thread:Terminate()
			ctx.mount_thread = nil
		end

		if ctx.status == "mounted" then
			os.execute( "/usr/bin/disk-dettach.lua " .. ctx.mount )
			local mounter = require "nas_NFS"
			if ctx.type and ctx.type ~= "" and ctx.type ~= "nfs" then
				if ctx.type == "cifs" then
					mounter = require "nas_CIFS"
				end
			end
			mounter.UMount(_G.g_event, id)
		end
		nas.mounts[id] = nil
	end

	cfg[id] = nil
	cfg:__TRIGGER(true)

	return true
end

function nas.ChooseNas(id, choose)
	local ctx = nas.mounts[id]
	if not ctx then
		return false, "Invalid id"
	end

	ctx.choose = choose
	cfg[id].choose = choose
	cfg:__TRIGGER(true)
	return true
end

function nas.Init()
	coaf.ScheduleDelayedTask(function()	
		local k,v
		for k,v in pairs(cfg) do
			nas.StartMount(k)
		end
	end)
end

nas.Rule = {
	ChooseNas = {
		is_open = true,
		parameters = {
            {key = "id", type = "string", must = true},
			{key = "choose", type = "boolean", must = true},
        }
	},
	Add = {
        is_open = true,
        parameters = {
            {key = "id", type = "string", must = true},
			{key = "name", type = "string", must = true},
			{key = "type", type = "string", must = true},
			{key = "host", type = "string", must = true},
			{key = "volumn", type = "string", must = true},
			{key = "options", type = "string", must = true},
        }
    },
	Remove = {
        is_open = true,
        parameters = {
            {key = "id", type = "string", must = true},
        }
	},
	GetList = {
        is_open = true,
        parameters = {}
	}
}

------------------------------------------------

coaf.Init( "nas_manager" )
-- Initialize service.
service.InitSocket()

---------------------------------------------------------
local type_change = {
    number = tonumber,
    string = tostring
}
local function empty_function() print("empty function") return true end
local function errror_function() print("error function") return false end
local function createResult(...)
    local result = {...}
    local is_ok = result[1]
    if not is_ok then
        print("###########XPCALL-ERROR#######", result[2])
        return
    end
    return table.unpack(result, 2, #result)
end

local function createNewObj(source, name)
    if not source or type(source) ~= "table" then
        return
    end
    
    return setmetatable({}, {
        __index = function(_, key)
            local func = source[key]
            if not func then
                print("not have this func and return", name)
                return empty_function
            end

            return function(...)
                local arguments, is_dict
                local tbl = {...}
                if type(tbl[1]) == "table" then
                    is_dict = true
                    arguments = tbl[1]
                else
                    arguments = tbl
                end
            
                local rule = source.Rule[key]
                -- print("key(function name) = ", key, name)
                if not rule then
                    print("not this rule and return empty function", name)
                    return empty_function()
                end
            
                if not rule.is_open then
                    --NOTE: 这里应该区别是网络请求还是内部coaf调用，如果是网络请求就return
                    print("not open this function: ", key, name)
                    return empty_function()
                end
            
                local args_list = {}
                for i, v in ipairs(rule.parameters) do
                    local value
                    if arguments[v.key] ~= nil then
                        value = arguments[v.key]
                    else
                        value = arguments[i]
                    end
                    if v.must and value == nil then
                        print("not have the parameters : ", v.key, name)
                        return errror_function()
                    end
            
                    if value ~= nil then
                        local change_func = type_change[v.type]
                        if change_func then
                            value = change_func(value)
                        end
                        if type(value) ~= v.type  then
                            print("parameters type is not allowed : ", v.key, name)
                            return errror_function()
                        end
                    end
            
                    args_list[i] = value
                end
            
                if is_dict then
                    return createResult(xpcall(function()
                        return source[key](table.unpack(args_list, 1, #rule.parameters))
                    end, debug.traceback))
                else
                    return createResult(xpcall(function()
                        return source[key](table.unpack(tbl, 1, #tbl))
                    end, debug.traceback))
                end
            end
        end,
    })
end

local function Main(args)
	_G.g_event = event.createNew( "nas_manager", true ) -- Don't start slave thread, since we never use it.
	g_service = service.New( "nas_manager", "unix:/var/run/nas.sock" )
	if g_service then
		g_service:CreateCoServerProxy()
		g_service:ExposeObject("/nas", createNewObj(nas, "nas") )
	else
		_dbg.Error( "NAS", "Fail to create CoService for nas_manager" )
	end
	nas.Init()
end

coaf.Start(Main,arg)

_dbg.Print( "NAS", "nas_manager service is terminated." )

