package.path = package.path .. ";/usr/local/bin/network/simpleNetwork/?.lua;"
local coaf = require "coaf.core"
local _dbg = require "coaf.debug"
local thread = require "coaf.thread"
local pt = require "pl.pretty" 

local is_ok, FEATURES = pcall(require, "features")

require "SimpleNetHelper"

local _M = {
    configuredIntfs = {},
	interfaceStatus = {}
}

local config = require "netcfg"
local funcs = require "nm_funcs"
local static = require "nm_static"
local hostap_m
if is_ok and FEATURES and FEATURES.FEATURE_SUPPORT_WIFI then
	hostap_m = require "hostap_m"
end
local is_ok, special = pcall(require , "nm_special")
if not is_ok then
    special = nil
end

local nm_config = config.Load("/etc/network/nm.default")

config.PATCHING = true
if special and special.PatchConfigs then
    nm_config = special.PatchConfigs(nm_config) or nm_config
end
config.PATCHING = false

local function GetDNSList()
	local f = io.open( "/etc/resolv.conf", "r" )
	if not f then
		return nil
    end
    local rslt = {}
    local l = f:read("*l")
    local addr
    while l do
        addr = l:match("^%s*nameserver%s+([%d%.]+)")
        if addr then
            table.insert(rslt, addr)
        end
        l = f:read("*l")
    end
    f:close()
    return rslt
end

local function checkInterfaces()
	if check_is_working then
		return
	end

	check_is_working = true
	local is_ok, err = xpcall( function()
		local intfs = funcs.EnumNetworks()
		if special and special.FixEnumeratedInterfaces then
			local _tmp_r = special.FixEnumeratedInterfaces( _M, intfs )
			if _tmp_r then
				intfs = _tmp_r
			end
		end

		local new_or_update_intfs = {}
		--Check carriers
		for i,v in ipairs(intfs) do
			if not _M.configuredIntfs[ v ] then
				_dbg.Info( "NETWORK", "New interface:", v, "found, configure it." )
				table.insert( new_or_update_intfs, v )
			elseif _M.interfaceStatus and _M.interfaceStatus[v] ~= intfs[v] then --Carrier changed.
				if intfs[v] and intfs[v] ~= 0 then
					_dbg.Info( "NETWORK", "Interface:", v, "carrier on, update the network configuration." )
				else
					_dbg.Info( "NETWORK", "Interface:", v, "carrier off, disable it." )
				end
				table.insert( new_or_update_intfs, v)
			end
		end

		_M.ConfigNetwork( new_or_update_intfs )

		for i,v in pairs(_M.configuredIntfs) do
			if not table.index( intfs, i ) then
				_M.configuredIntfs[i] = nil -- Interface lost/no carrier, drop it.
			end
		end

		_M.interfaceStatus = intfs
	end, debug.traceback ) --End of pcall.

    if not is_ok then
        _dbg.Error("NETWORK", "Do xpcall failed, err :", err)
    end

	check_is_working = false
end

local preiodCheck, hostapCheck, hostapStateCheck
--NOTE: 通用接口 初始化模块
function _M.Init(coservice)
	if special and special.LoadDrivers then
		special.LoadDrivers()
	end
	static.Init()
	_M.Reset()
	if hostap_m then
		coaf.ScheduleDelayedTask(function()
			hostap_m.Init()
		end, 5)

		hostapCheck = coaf.NewTimer(hostap_m.AutoConnectSaved, 10, 5):start()
		hostapStateCheck = coaf.NewTimer(hostap_m.UpdateWpaState, 11, 5):start()
	end
	
    if coservice then
        -- coservice:ExposeObject("/config/nm", nm_config)
		coservice:ExposeObject("/static", createNewObj(static))
		if hostap_m then
			coservice:ExposeObject("/hostap", createNewObj(hostap_m))
		end
    end

    preiodCheck = coaf.NewTimer(checkInterfaces, 1, 2)
    if preiodCheck then
        preiodCheck:start()
    end
end

--NOTE: 通用接口 获取所有设备
function _M.GetLinkList()
    return SnH.GetAllInterfaces()
end

--NOTE: 通用接口 获取某个设备的具体信息
function _M.GetLinkDetail(ifname)
	ifname = ifname or "eth0"
    if not nm_config.ethernets and nm_config.ethernets[ifname] then
        return
    end

    local cfg = nm_config.ethernets[ifname]
    local ret = funcs.GetLinkDetail(ifname)
    ret.method = cfg.method
	ret.aliases = nm_config.aliases[ifname]
    if cfg.method ~= "dhcp" then
        if not ret.address or ret.address == "0.0.0.0" then
            ret.address = cfg.address
            ret.netmask = cfg.netmask
        end
        if not ret.gw or ret.gw == "" or ret.gw == "0.0.0.0" then
            ret.gw = cfg.gw
        end
        ret.dns = cfg.dns
    else
        ret.dns = GetDNSList()
    end
	ret.dns = table.concat(ret.dns, ",")
    return ret
end

--NOTE: 通用接口 获取某个设备的状态  是否重复？
function _M.GetLinkStatus(ifname)
    return funcs.GetLinkStatus(ifname)
end

--NOTE: 通用接口， 配置所有网络？？？  并没有在api 层 发现哪里调用这个方法， 是否可以去掉？
function _M.ConfigNetwork(intfs)
	local all
    if not intfs then
        intfs = _M.GetLinkIP()
		all = intfs
    end

    local r, err
    if nm_config.interlink then
		r = table.index( intfs, nm_config.interlink.ifname )
		--local carrier = intfs[nm_config.interlink.ifname]
		if r then
			table.remove( intfs, r )
			_M.configuredIntfs[ nm_config.interlink.ifname ] = true -- Mark it configured.
			r, err = funcs.ConfigInterlink( nm_config.interlink )
			if not r then
				_dbg.Error( "NETWORK", "Fail to configure inter-link, with error:", err )
			end
		end
	end

    for i, ifname in ipairs(intfs) do
		_M.configuredIntfs[ ifname ] = true -- Mark it configured.

		if nm_config.ethernets then
			local cfg = nm_config.ethernets[ifname]
			if not cfg then
				nm_config:S_ADD( ifname, {} ) -- Not configured, add it.
				cfg = nm_config.ethernets[ifname]	--Default config
			end

			r, err = funcs.ConfigEthernet( ifname, cfg )
			if not r then
				_dbg.Error( "NM", "Fail to configure interface of '" .. ifname .. "', error:", err )
			end
		end

		if nm_config.aliases then
			local cfg = nm_config.aliases[ifname]
			if cfg then
				r, err = funcs.ConfigEthernetAlias( ifname, cfg )
				if not r then
					_dbg.Error( "NM", "Fail to configure interface alias of '" .. ifname .. "', error:", err )
				end
			end
		end

		all = all or _M.GetLinkIP()
		if all[ifname] == 1 then
			static.Start(ifname)
		end

		if special and special.InterfaceSpecialConfig then
			special.InterfaceSpecialConfig( _M, ifname, intfs[ifname] and intfs[ifname]~=0 or false )
		end
	end -- for.
end

--NOTE: 通用接口 枚举网络接口？
function _M.GetLinkIP()
    local intfs = funcs.EnumNetworks()
	if special and special.FixEnumeratedInterfaces then
        intfs = special.FixEnumeratedInterfaces( _M, intfs ) or intfs
	end
	return intfs
end

--NOTE: 通用接口 设置某个有线连接
function _M.ChangeConnection(ifname, cfg)
	ifname = ifname or "eth0"
	if type(cfg.dns) == "string" then
		cfg.dns = split(cfg.dns, ",")
	end
	table.update(nm_config.ethernets[ifname], cfg)
	if cfg.dns then
		nm_config.ethernets[ifname].dns = cfg.dns
	end
	table.update(nm_config.aliases[ifname], cfg.aliases)
	nm_config:__TRIGGER(true)
	return true
end

function _M.StartMDNS()
    if nm_config.enable_mdns and special and special.StartAvahiDaemon then
		local ignoreAddrs = {}
		if nm_config.aliases then
			local k,v
			for k,v in pairs(nm_config.aliases) do
				if v.enable then
					if v.address and v.address ~= "" and v.address ~= "0.0.0.0" then
						table.insert( ignoreAddrs, v.address )
					end

					if v.address2 and v.address2 ~= "" and v.address2 ~= "0.0.0.0" then
						table.insert( ignoreAddrs, v.address2 )
					end
				end
			end
		end

		if  nm_config.interlink 
			and nm_config.interlink.enable 
			and nm_config.interlink.address 
			and nm_config.interlink.address ~= "" 
			and nm_config.interlink.address ~= "0.0.0.0" then
			table.insert( ignoreAddrs, nm_config.interlink.address )
		end

		special.StartAvahiDaemon( ignoreAddrs )
	end
end

--NOTE: MDNS的开启关闭接口可能不需要提取出来，因为simpleNetworkmanager中没有具体实现
--NOTE: 通用接口 停止AVAHI-Daemon(DNS),这里是用avahi实现的dns？ special目前是个空表，这里应该没有任何意义
function _M.StopMDNS()
	if nm_config.enable_mdns and special and special.StopAvahiDaemon then
		special.StopAvahiDaemon()
	end
end

--NOTE: MDNS的开启关闭接口可能不需要提取出来，因为simpleNetworkmanager中没有具体实现
--NOTE: 通用接口 这里是用于复位？
function _M.Reset()
    if special and special.BeforeReset then
		special.BeforeReset( _M )
	end
	_M.StopMDNS()
	_M.configuredIntfs = {}
	_M.ConfigNetwork()
	_M.StartMDNS()
	if special and special.AfterReset then
		special.AfterReset( _M )
	end
end

--NOTE: 通用接口 释放资源
function _M.Exit()
    if preiodCheck then
        preiodCheck:destroy()
    end

	if hostapCheck then
		hostapCheck:destroy()
	end

	if hostapStateCheck then
		hostapStateCheck:destroy()
	end

    if special and special.RemoveDrivers then
        special.RemoveDrivers()
    end

	if hostap_m then
		hostap_m:Destroy()
	end
end

_M.doc = ""
_M.Rule = {
	GetLinkList = {
		parameters = {},
		is_open = true
	},
	GetLinkDetail = {
		parameters = {
			{key = "ifname", type = "string"},
		},
		is_open = true
	},
	ChangeConnection = {
		parameters = {
			{key = "ifname", type = "string"},
			{key = "cfg", type = "table", must = true}
		},
		is_open = true
	},
}

return _M