local coaf = require "coaf.core"
local lfs = require "lfs"
local _dbg = require "coaf.debug"
local thread = require "coaf.thread"
local run_daemon = require "coaf.run_daemon"
pcall(require, "features")

local os_date = os.date
local string_format = string.format
local Gprint = _G.print
local print = function(...)
	local now = os_date("*t")
	local tbl = {...}
	table.insert(tbl, string_format("%02d:%02d:%02d", now.hour, now.min, now.sec))
	Gprint(table.unpack(tbl))
end

local client = require "coaf.client"
local adapter = require "coaf.cli_adapter.simple"
client.InitSocket(nil, adapter)

local function GetProxy(objPath, service, loc, timeout, socket_addr)
	local proxy
	local cli

	cli = client.New( service, loc, socket_addr, timeout or 5 ) --5 seconds timeout.
	if not cli then
		_dbg.Error( "CO", "Fail to create connection to service '" .. (loc or "local") .. "/" .. (service or "nil") )
		return nil
	end
	proxy = cli:CreateProxy(objPath, true) --Exclusive
	if not proxy then
		_dbg.Error( "CO", "Fail to create object proxy for '" .. (objPath or "unknown") .. "'" )
		return nil
	end
	return proxy
end

local function startCoafRunDaemon(objPath, service, args)
	table.insert(args, 1, objPath)
	table.insert(args, 1, service)
	table.insert(args, 1, "local")
	local task = run_daemon.Create({
		name = "CoafRundaemon:" .. string.format("objPath=%s,service=%s,func=%s", objPath, service, args[4]),
		run = "/usr/local/bin/cocall.lua",
		args = args
	})
	if task then
		task:Start()
	end
end

-------------------------

local _M = {
	Identifier = "RECORD_MAN",
	FriendlyName = "Disk and record Manager",
	Version = "2.0",
	ConfigPrefix = "record_m",

	ALL_DISKS = {},	--Record all disks and them's mounts.

	RECORDING = {
		[0] = {},
		[1] = {}
	},

	SEQUENCE_COUNTER = 0,
}

local StreamsMap = {
	[0] = "main",
	[1] = "sub"
}

-------------------------

----------------------------------------------------
-- General module interface
----------------------------------------------------

function _M.Activate()
	--Nothing
end

function _M.Start()
end

function _M.Stop()
end

function _M.Reset()
--TODO: Nothing
end

function _M.Terminate()
	os.exit(0)
end

function _M.Control(action,params)
	if action == "activate" then
		_M.Activate()
		return 0, {}
	elseif action == "start" then
		_M.Start()
		return 0, {}
	elseif action == "stop" then
		_M.Stop()
		return 0, {}
	elseif action == "Reset" then
		_M.Reset()
		return 0, {}
	elseif action == "terminate" then
		_M.Terminate()
		return 0, {}
	else
		return -1, {}
	end
end

----------------------------------------------------
-- General configure interface
----------------------------------------------------

function _M.GetInfo()
	local cfg = {}
	for k, v in pairs(_G.record_m[0]) do
		cfg[k] = v
	end
	cfg.start = nil
	return cfg
end

function _M.Update(cfg)
	if cfg.start ~= nil then
		if not cfg.start then
			_M.StopRecord("main")
			coaf.ScheduleDelayedTask(function()
				_M.StopRecord("sub")
			end, 0.5)
		else
			_M.StartRecord("main")
			coaf.ScheduleDelayedTask(function()
				_M.StartRecord("sub")
			end, 0.5)
		end
	end
	table.update(_G.record_m[0], cfg)
	table.update(_G.record_m[1], cfg)
	_G.record_m:__TRIGGER(true)
	return true
end

function _M.RestoreDefault()
	--TODO:
	return 0
end

local function GetDevFilesystem(device)
	if not device or device == "" then
		return
	end

	local f = io.popen("/sbin/fdisk -l | grep '/dev/" .. device .. "'")
	if not f then
		return
	end

	local filesys
	local filesys_s = f:read("l")
	f:close()

	if filesys_s:match("FAT") then filesys = "vfat" end
	if filesys_s:match("NTFS") then filesys = "ntfs" end
	return filesys
end

----------------------------------------------------
-- Disk interface
----------------------------------------------------

local checkReady
local recordReadyStartDict = {}
function _M.RecordStateCheck()
	for k, v in pairs(recordReadyStartDict) do
		if v == true then
			coaf.ScheduleDelayedTask(function()
				_M.StartRecord(k)
			end, k == "main" and 0.01 or 0.5)
		else
			coaf.ScheduleDelayedTask(function()
				_M.StopRecord(k)
			end, k == "main" and 0.01 or 0.5)
		end
		recordReadyStartDict[k] = nil
	end
	checkReady = nil
end

function _M.DiskAttach(device, type, volName, filesys)
	local filesys = filesys or GetDevFilesystem(device)
	if not device or device == "" or not type or type == "" or not filesys or filesys == "" then
		print( "ERROR: Disk attach: invalid arguments" )
		return
	end

	print( "A disk is attached: DEVICE='" .. device .. "', TYPE='" .. type .. "', FILE_SYS='" .. filesys .. "', VOLUMN_NAME='" .. (volName or "") .. "'" )
	
	if _M.ALL_DISKS[device] then
		print( "Current device is already attached, so remount it. (A manually dettach should occus)" )
		_M.DiskDettach(device)
	end

	print( "Mounting device..." )
	local r, msg = _M.MountDisk( device, type, filesys )
	if not r then
		print( "Mount ERROR: ", msg )
		return
	else
		print( "Mount DONE.")
	end

	--Attached, so initialize the recording database.
	_M.InitRecordDB( device )
	_M.ALL_DISKS[device].VOLNAME = volName == "" and device or volName
	local done, total, used, free, percent = _M.GetDiskSpaces( device, _M.ALL_DISKS[device].TYPE == "nas" )
	if done then
		print( "Disk '" .. device .. "' info: TOTAL=" .. total .. " MB, USED=" .. used .. " MB, FREE=" .. free .. " MB, Percent=" .. percent .. "%" ) 
		_M.ALL_DISKS[device].TOTAL = total
		_M.ALL_DISKS[device].USED = used
		_M.ALL_DISKS[device].FREE = free
		_M.ALL_DISKS[device].PERCENT = percent
	else
		_M.ALL_DISKS[device].TOTAL = 0
		_M.ALL_DISKS[device].USED = 0
		_M.ALL_DISKS[device].FREE = 0
		_M.ALL_DISKS[device].PERCENT = 100
	end

	--Recording is disabled.
	if _G.DISABLE_RECORD then
		return
	end

	for k, v in pairs(StreamsMap) do
		if _G.record_m[k].auto_record and type ~= "nas" then	--Auto start recording, so try start it
			_G.record_m[k].start = true
			recordReadyStartDict[v] = true
			if not checkReady then
				checkReady = true
				thread.NewTimer(_M.RecordStateCheck, 2):start()
			end
		end
	end
	_G.record_m:__TRIGGER(true)
end

function _M.DiskDettach(device)
	print("\n\n========== in function DiskDettach =======\n\n")
	print( "A disk is dettached: DEVICE='" .. device .. "'" )

	for k,v in pairs( _M.RECORDING) do
		if (v.Status == "starting" or v.Status == "started" or v.Status == "stopping") then
			local skip
			for _, vv in ipairs(v.NasDevice) do
				if vv == device then
					skip = true
					break
				end
			end
			if skip then
				print( "Recording stream '" .. StreamsMap[k] .. "' is depended by this device, so the recording should be stopped." )
				_G.record_m[k].start = false
				recordReadyStartDict[StreamsMap[k]] = false
				if not checkReady then
					checkReady = true
					thread.NewTimer(_M.RecordStateCheck, 2):start()
				end
				-- _M.StopRecord(StreamsMap[k])
			end
		end
	end

	_M.UmountDisk(device)
	print( "Disk is un-mounted." )
end

function _M.DiskList()
	local proxy = GetProxy("/nas", "nas_manager")
	local data = {}
	if proxy then
		local is_ok, x = proxy.GetList()
		if is_ok then
			data = x
		end
		proxy:Destroy()
	end
	local result = { usb = {}, nas = {} }
	for k, v in pairs(_M.ALL_DISKS) do
		v.key = k
		if v.TYPE == "nas" then
			for idx = #data, 1, -1 do
				local one = data[idx]
				if one.mount == k then
					v.choose = one.choose
					v.host = one.host
					v.status = one.status
					v.id = one.id
					v.VOLNAME = one.name
					table.remove(data, idx)
					break
				end
			end
			table.insert(result.nas, v)
		else
			table.insert(result.usb, v)
		end
	end

	for _, v in ipairs(data) do
		v.MOUNT = v.host
		v.VOLNAME = v.name
		table.insert(result.nas, v)
	end

	table.sort(result.usb, function(a, b)
		return a.MOUNT > b.MOUNT
	end)

	table.sort(result.nas, function(a, b)
		return a.MOUNT > b.MOUNT
	end)
	return result
end

function _M.DiskDetail(device)
	local r = _M.ALL_DISKS[device]
	if not r then
		return -1, {}
	else
		return 0, r
	end
end

function _M.DiskRecordList(device,filter)
	if not device then
		return -1, {{}}
	end

	local r = _M.ALL_DISKS[device]
	if not r then
		return -1, {{}}
	else
		local rr = _M.CheckRecordList( device, filter )
		if not rr then
			return -1, {{}}
		else
			return #rr, rr
		end
	end
end

----------------------------------------------------
-- Record interface
----------------------------------------------------

function _M.RegisterWorker( stream )
	local streamIdx
	if not stream or stream == "any" then
		streamIdx = -1
	elseif stream == "main" or stream == "0" then
		streamIdx = 0
	elseif stream == "sub" or stream == "1" then
		streamIdx = 1
	else
		streamIdx = -1
	end

	for k, v in pairs( _M.RECORDING) do
		if streamIdx == -1 or streamIdx == k then
			if v.Status == "starting" then	--Pending, that 'streamer' module may not get the start message, so try again.
				_M.StartRecord( k == 0 and "main" or "sub" )
			elseif v.Status == "started" then --Why a started job but new worker registered? I think the worker may crashed and start again!
											  --So I need stop current first, then try start again!
				_M.StopRecord( k == 0 and "main" or "sub" )
				_M.StartRecord( k == 0 and "main" or "sub" )
			elseif v.Status == "stopping" then --Pending, that 'streamer' module may not get the stop message, so try again.
				_M.StopRecord( k == 0 and "main" or "sub" )
			end
		end
	end
	return true
end

function _M.AckStartRecord( sequence_id, ackCode )
	local k,v
	for k,v in pairs( _M.RECORDING) do
		if v.Status == "starting" and v.SequenceID == sequence_id then
			if ackCode == "done" or ackCode == "started" then
				print( "INFO: The recording sequence of '" .. sequence_id .. "' is acked started." )
				v.Status = "started"
				v.ErrorAction = ""
				v.Error = ""
			else
				print( "INFO: The recording sequence of '" .. sequence_id .. "' is acked start error: " .. ackCode )
				v.Status = "stopped"
				v.ErrorAction = "start"
				v.Error = ackCode or "unknown"
			end
			break
		end
	end
	--TODO: recording status.
end

function _M.AckStopRecord( sequence_id, ackCode )
	local k,v
	for k,v in pairs( _M.RECORDING) do
		if v.Status == "stopping" and v.SequenceID == sequence_id then
			if ackCode == "done" or ackCode == "stopped" then
				print( "INFO: The recording sequence of '" .. sequence_id .. "' is acked stopped." )
				v.Status = "stopped"
				v.ErrorAction = ""
				v.Error = ""
			else
				print( "INFO: The recording sequence of '" .. sequence_id .. "' is acked stop error: " .. ackCode )
				v.Status = "stopped"
				v.ErrorAction = "stop"
				v.Error = ackCode or "unknown"
			end

			v.CURRENT_FILE = nil
			v.CURRENT_START_TIME = nil
			v.CURRENT_END_TIME = nil
			v.START_TIME = nil
			v.END_TIME = nil
			v.FORMAT = nil
			v.FILES = nil
			v.action = nil
			collectgarbage()
			break
		end
	end
end

function _M.UpdateRecord( sequence_id, action, filename, format, time, size_kb )
	print( "INFO: Update recording: SequenceID='" .. sequence_id .. "', Action='" .. action .. "', Filename='" .. filename .. "' ..." )
	local k,v
	for k,v in pairs(_M.RECORDING) do
		-- If matched sequence-id and current is started.
		if v.Status == "started" and v.SequenceID == sequence_id then
			v.action = action
			if action == "new" then
				v.CURRENT_FILE = filename
				v.CURRENT_START_TIME = tonumber(time)
				v.START_TIME = v.START_TIME or tonumber(time)
				v.END_TIME = tonumber(time) or v.START_TIME
				v.CURRENT_END_TIME = v.END_TIME
				v.FORMAT = format or "ts"
				if not v.FILES then
					v.FILES = { [filename] = size_kb or 0 }
				else
					v.FILES[ filename ] = size_kb or 0
				end
			elseif action == "update" or action == "end" then
				v.CURRENT_FILE = filename --Just update it.
				v.FORMAT = format or v.FORMAT or "ts"
				v.START_TIME = v.START_TIME or tonumber(time)
				v.END_TIME = tonumber(time) or v.START_TIME
				v.CURRENT_END_TIME = v.END_TIME
				if not v.FILES then
					v.FILES = { [filename] = size_kb or 0 }
				else
					v.FILES[ filename ] = size_kb or 0
				end
			elseif action == "stop" or action == "terminate" then
				_G.record_m[k].start = false
				_M.AckStopRecord( sequence_id, "stopped" )
				return "ok"
			end
			--TODO: Update the meta file.
			-- coaf.ScheduleDelayedTask(function()
				for _, disk in ipairs(v.NasDevice) do
					local diskMount = _M.ALL_DISKS[disk].MOUNT
					local file
					os.execute( "/bin/mkdir -p " .. diskMount .. "/RECORD/" .. sequence_id )
					file = io.open( diskMount .. "/RECORD/" .. sequence_id .. "/.START", "w" )
					if file then
						file:write( v.START_TIME .. "\n" )
						file:close()
					end
					
					file = io.open( diskMount .. "/RECORD/" .. sequence_id .. "/.END", "w" )
					if file then
						file:write( v.END_TIME .. "\n" )
						file:close()
					end
	
					file = io.open( diskMount .. "/RECORD/" .. sequence_id .. "/.FORMAT", "w" )
					if file then
						file:write( v.FORMAT .. "\n" )
						file:close()
					end
				end
			-- end)
			return "ok"
		elseif v.SequenceID == sequence_id then
			if action == "stop" or action == "terminate" then
				_G.record_m[k].start = false
				_M.AckStopRecord( sequence_id, "stopped" )
				return "ok"
			end

			print( "***WARNING: A recording task is not 'started' but '" .. v.Status .. "', but record update message received!" )
			if v.Status ~= "stopped" and v.Status ~= "stopping" then
				v.Error = "status-error"
				v.ErrorAction = "update-record"
			end
			return "error"
		end
	end
	
	print( "***WARNING: Invalid sequence-id but record update message received!" )
	return "error"
end

function _M.GetStatus(stream)
	local streamIdx = 0
	if stream == "sub" or tonumber(stream) == 1 then
		streamIdx = 1
	end

	if _M.RECORDING[streamIdx].Status ~= "started" then
		if not _M.RECORDING[streamIdx].Status then
			-- return "idle",{}
			return {status = "idle"}
		else

			if _M.RECORDING[streamIdx].Error and _M.RECORDING[streamIdx].Error ~= "" then
				-- return "error:" .. (_M.RECORDING[streamIdx].ErrorAction or "") .. ":" .. _M.RECORDING[streamIdx].Error, {}
				return {status = "error", action = _M.RECORDING[streamIdx].ErrorAction or "", error = _M.RECORDING[streamIdx].Error}
			else
				-- return _M.RECORDING[streamIdx].Status, {}
				return {status = _M.RECORDING[streamIdx].Status}
			end
		end
	else
		local allitems = _M.CheckRecordList( _M.RECORDING[streamIdx].NasDevice, _M.RECORDING[streamIdx].SequenceID )
		for _, items in pairs(allitems) do
			if items and items[1] then
				local v = _M.RECORDING[streamIdx]
				items[1].CURRENT_FILE = v.CURRENT_FILE or ""
				items[1].CURRENT_START_TIME = v.CURRENT_START_TIME or 0
				items[1].CURRENT_END_TIME = v.CURRENT_END_TIME or 0
			end
		end
		-- return "started", items and items[1] or {}
		return {status  = "started"}
	end
end

----------------------------------------------------
-- Disk mount/umount
----------------------------------------------------
function _M.MountDisk( device, type, filesys)
	if not device or not type or not filesys then
		return false, "Invalid arguments"
	end

	if  _M.ALL_DISKS[device] then
		return false, "Already mount"
	end

	if type == "nas" then --Since NAS is already mounted.
		_M.ALL_DISKS[device] = {
			DEVICE = deivce,
			TYPE = type,
			FILESYS = filesys,
			MOUNT = device
		}
		return true
	end

	local mountPath = "/media/" .. type .. "/" .. device
	local ret, status, code = os.execute( "/bin/mkdir -p " .. mountPath  )
	if not ret or status ~= "exit" or code ~= 0 then
		return false, "Fail to create mount point"
	end

	if filesys == "ntfs" then
		ret, status, code = os.execute( "/usr/bin/ntfs-3g -o utf8,big_writes,gid=100,umask=0 /dev/" .. device .. " " .. mountPath )
		if not ret or status ~= "exit" or code ~= 0 then
			--Try umount then mount, try it again.
			print( "*** First mount chance is fail. Try umount first, then try again." )
			os.execute( "/bin/umount -l /dev/" .. device )

			ret, status, code = os.execute( "/usr/bin/ntfs-3g -o utf8,big_writes,gid=100,umask=0 /dev/" .. device .. " " .. mountPath )
			if not ret or status ~= "exit" or code ~= 0 then
				return false, "Fail to mount NTFS device at '" .. mountPath .. "'"
			end
		end
	else
		ret, status, code = os.execute( "/bin/mount -t " .. filesys .. " -o utf8,gid=100,umask=0 /dev/" .. device .. " " .. mountPath )

		if not ret or status ~= "exit" or code ~= 0 then
			--Try umount then mount, try it again.
			print( "*** First mount chance is fail. Try umount first, then try again." )
			os.execute( "/bin/umount -l /dev/" .. device )
			ret, status, code = os.execute( "/bin/mount -t " .. filesys .. " -o utf8,gid=100,umask=0 /dev/" .. device .. " " .. mountPath )
			if not ret or status ~= "exit" or code ~= 0 then
				return false, "Fail to mount " .. filesys .. " device at '" .. mountPath .. "'"
			end
		end
	end

	_M.ALL_DISKS[device] = {
		DEVICE = deivce,
		TYPE = type,
		FILESYS = filesys,
		MOUNT = mountPath
	}
	return true
end

----------------------------------------------------

function _M.UmountDisk( device )
	if not device then
		return false
	end

	local diskItem = _M.ALL_DISKS[device]
	if not diskItem then
		return true
	end

	if diskItem.type ~= "nas" then
		os.execute( "/bin/umount -l " .. diskItem.MOUNT )
	end

	_M.ALL_DISKS[device] = nil
	return true
end

----------------------------------------------------

function _M.GetDiskSpaces( device, isNAS )
	local file
	if isNAS then
		file = io.popen("/bin/df -m " .. device, "r")
	else
		file=io.popen("/bin/df -m /dev/" .. device,"r")
	end

	if not file then
		return false, "Fail to get disk spaces"
	end

	local total, used, free, percent
	local line = file:read("l")
	while line do
		local flag,t,u,f,p = line:match( "(.-)%s+(%d+)%s+(%d+)%s+(%d+)%s+(%d+)%%%s+" )
		if flag then
			total = tonumber(t)
			used = tonumber(u)
			free = tonumber(f)
			percent = tonumber(p)
			break
		end
		line=file:read("l")
	end
	file:close()

	if total then
		return true, total, used, free, percent
	else
		return false, "Fail to get disk spaces"
	end
end

----------------------------------------------------

function _M.Init()
	run_daemon:Init()
	-- _M.codecProxy = GetProxy("/record", "codec")
end

function _M.PickNas()
	local list = {}
	local proxy = GetProxy("/nas", "nas_manager")
	if proxy then
		local is_ok, data = proxy.GetList()
		proxy:Destroy()
		if is_ok then
			for _, v in ipairs(data or {}) do
				if v.choose and _M.ALL_DISKS[v.mount] then
					table.insert(list, v.mount)
				end
			end
		end
	end
	return list
end

local function ChooseFirstDist()
	local list = {}
	for k in pairs(_M.ALL_DISKS) do
		table.insert(list, k)
	end
	table.sort(list, function(a, b)
		return a < b
	end)
	return list[1]
end

----------------------------------------------------
function _M.PickDisk( req, choose )
	if not choose then
		return
	end
	req = req or "largest"
	
	local reqType = "usb"
	if req:match("sdmmc") then
		reqType = "sdmmc"
	elseif req:match( "usb" ) then
		reqType = "usb"
	elseif req:match( "nas" ) then
		reqType = "nas"
	end

	local reqMatch = "largest"
	if req:match( "first" ) then
		reqMatch = "first"
	elseif req:match( "mark" ) then
		reqMatch = "mark"
	end

	local shouldBe = {}
	local shouldCount = 0
	local largestDev
	local largestSize = 0
	local k,v
	for k,v in pairs( _M.ALL_DISKS ) do
		if not reqType or reqType == v.TYPE then
			if reqMatch == "first" then
				return ChooseFirstDist()
			elseif reqMatch == "largest" then
				if not largestDev or v.FREE > largestSize then
					largestDev = k
					largestSize = v.FREE
				end
			else
				shouldBe[k] = v.MOUNT
				shouldCount = shouldCount + 1
			end
		end
	end

	if reqMatch == "largest" then
		return largestDev
	elseif reqMatch == "mark" then
		if shouldCount == 0 then
			return nil
		else
			local firstOne
			local file

			for k,v in pairs(shouldBe) do
				if not firstOne then
					firstOne = k
				end

				file = io.open( v .. "/RECORD_HERE.txt", "r" )
				if file then
					file:close()
					return k
				end
				file = io.open( v .. "/RECORD_HERE", "r" )
				if file then
					file:close()
					return k
				end
			end
			return firstOne
		end
	else
		return nil
	end
end
----------------------------------------------------

function _M.StartRecord( stream )
	if not stream then
		print( "ERROR: Start recording with invalid arguments or no valid DBUS service." )
		return false
	end

	local streamIdx = 0
	local streamFlag = "M"

	stream = stream:lower()
	if stream == "sub" or stream == "1" then
		stream = "sub"
		streamFlag = "S"
		streamIdx = 1
	else
		stream = "main"
	end

	local recordItem = _M.RECORDING[streamIdx]
	if not recordItem then
		return false
	end

	local alreadyStart = false
	if recordItem.Status == "starting" or recordItem.Status == "started" then
		print( "Start record [" .. stream .. "]: Recording is already started, just update it. Sequence-ID is [" .. recordItem.SequenceID .. "]" )
		alreadyStart = true
		recordItem.ErrorAction = ""
		recordItem.Error = ""
	else
		local now = os.date("*t")
		local nowStr = string.format("%04d%02d%02d_%02d%02d%02d_%04d", now.year, now.month, now.day, now.hour, now.min, now.sec, _M.SEQUENCE_COUNTER)
		recordItem.SequenceID = streamFlag .. "-" .. nowStr
		_M.SEQUENCE_COUNTER = _M.SEQUENCE_COUNTER + 1
		print( "Start record [" .. stream .. "]: Prepair to start recording, *NEW* Sequence-ID is [" .. recordItem.SequenceID .. "]" )
		recordItem.Status = "starting" -- Start pending.
		recordItem.ErrorAction = ""
		recordItem.Error = "" --Clear error status.
		recordItem.DiskDevice = nil	--Drop saved, if exists.

		recordItem.CURRENT_FILE = nil
		recordItem.CURRENT_START_TIME = nil
		recordItem.CURRENT_END_TIME = nil

		recordItem.START_TIME = nil
		recordItem.END_TIME = nil
		recordItem.FORMAT = nil
		recordItem.FILES = nil
		recordItem.action = nil
	end

	recordItem.DiskPolicy = _G.record_m[streamIdx].disk_policy	--Just update.

	local format = _G.record_m[streamIdx].format
	if format == "" then
		format = "ts"
	end

	local prefix = _G.record_m[streamIdx].file_prefix .. "-" .. stream
	
	if recordItem.DiskDevice then
		--Check is it exist yet?
		if not _M.ALL_DISKS[recordItem.DiskDevice] then
			print( "WARNING: Current recording disk '" .. recordItem.DiskDevice .. "' is lost, try select another!" )
			recordItem.DiskDevice = nil
		end
	end

	if recordItem.NasDevice and next(recordItem.NasDevice) then
		for i = #recordItem.NasDevice, 1, -1 do
			if not _M.ALL_DISKS[recordItem.NasDevice[i]] then
				table.remove(recordItem.NasDevice, i)
			end
		end
	end

	recordItem.DiskDevice = recordItem.DiskDevice or _M.PickDisk( _G.record_m[streamIdx].disk_type, _G.record_m[streamIdx].choose )
	recordItem.NasDevice = _M.PickNas() or {}

	if not recordItem.DiskDevice and #recordItem.NasDevice == 0 then
		print( "ERROR: No valid disk, recording cannot be started." )
		if alreadyStart then
			_M.StopRecord( stream )
		end
		recordItem.Status = "stopped"
		recordItem.ErrorAction = "start"
		recordItem.Error = "nodisk"
		return false
	else
		print( "Selected recording disk: '" .. tostring(recordItem.DiskDevice) .. "' Nas: ", table.unpack(recordItem.NasDevice or {})  )
	end

	table.insert(recordItem.NasDevice, recordItem.DiskDevice)
	if not _M.ReserveDiskSpace( streamIdx, 300 ) then --Before record, must try keep 100MB spaces
		print( "ERROR: No disk space." )
		if alreadyStart then
			_M.StopRecord( stream )
		end
		recordItem.Status = "stopped"
		recordItem.ErrorAction = "start"
		recordItem.Error = "nospace"
		return false
	end
	-- local diskPath = _M.ALL_DISKS[recordItem.DiskDevice].MOUNT .. "/RECORD"

	for _, v in ipairs(recordItem.NasDevice) do
		local diskPath = _M.ALL_DISKS[v].MOUNT .. "/RECORD"
		os.execute( "/bin/mkdir -p " .. diskPath .. "/" .. recordItem.SequenceID )
	
		local limit_type = _G.record_m[streamIdx].limit_type
		local limit = 0
		if limit_type == "time" or limit_type == "time-loop" then
			limit = _G.record_m[streamIdx].limit_time
		elseif limit_type == "size" or limit_type == "size-loop" then
			limit = _G.record_m[streamIdx].limit_size
		else
			limit_type = "none"
		end
	
		-- coaf.ScheduleDelayedTask(function()
			print( "NOTIFY: Notify the recording is started.", stream )
			-- _M.codecProxy = _M.codecProxy or GetProxy("/record", "codec")
			-- if _M.codecProxy then
			-- 	_M.codecProxy.NotifyStartRecord( _G.record_m[streamIdx].bindAudio, _M.ALL_DISKS[v].FILESYS, stream, recordItem.SequenceID, format, diskPath, prefix, limit_type, limit)
			-- end
			startCoafRunDaemon("/record", "codec", {"NotifyStartRecord", _G.record_m[streamIdx].bindAudio, _M.ALL_DISKS[v].FILESYS, stream, recordItem.SequenceID, format, diskPath, prefix, limit_type, limit})
		-- end, stream == "main" and 0.01 or 0.2)
	end
	return true
end

----------------------------------------------------

function _M.StopRecord( stream, isall )
	if not stream then
		print( "ERROR: Stop recording with invalid arguments or no valid DBUS service." )
		return false
	end

	local streamIdx = 0
	stream = stream:lower()
	if stream == "sub" or stream == "1" then
		stream = "sub"
		streamIdx = 1
	else
		stream = "main"
	end

	local recordItem = _M.RECORDING[streamIdx]
	if not recordItem then
		return false
	end

	local doStop
	if recordItem.Status == "starting" or recordItem.Status == "started" then
		print( "Stop record [" .. stream .. "]: Sequence-ID is [" .. ( recordItem.SequenceID or "nil" ) .. "]" )
		recordItem.Status = "stopping" --stopping
		doStop = true
	elseif recordItem.Status ~= "stopping" then	--If stopping, need notify again.
		print( "Stop record [" .. stream .. "]: Already stopped." )
		return true
	end

	-- coaf.ScheduleDelayedTask(function()
	-- 	_M.codecProxy = _M.codecProxy or GetProxy("/record", "codec")
	-- 	if _M.codecProxy then
	-- 		_M.codecProxy.NotifyStopRecord( stream, recordItem.SequenceID or "" )
	-- 	end
	-- end, stream == "main" and 0.01 or 0.2)
	startCoafRunDaemon("/record", "codec", {"NotifyStopRecord", stream, recordItem.SequenceID or ""})
	return true, doStop
end

----------------------------------------------------
function _M.InitRecordDB( device )
	if not device then
		print( "ERROR: Invalid disk device while initialize recording." )
		return false
	end

	local mountPoint = _M.ALL_DISKS[device].MOUNT
	if not mountPoint then
		print( "ERROR: Invalid disk mounts while initialize recording." )
		return false
	end

	local path = mountPoint .. "/RECORD"
	local ret, status, code = os.execute( "/bin/mkdir -p " .. path )
	if not ret or status ~= "exit" or code ~= 0 then
		print( "ERROR: Fail to initialize recording database." )
		return false
	end
	return true
end
----------------------------------------------------
function _M.CheckRecordList( device, filter )
	local alldevice
	if type(device) ~= "table" then
		alldevice = {device}
	else
		alldevice = device
	end

	local device_result = {}
	for _, device in pairs(alldevice) do
		if not _M.InitRecordDB(device) then
			return nil
		end
	
		local mountPoint = _M.ALL_DISKS[device].MOUNT
		local path = mountPoint .. "/RECORD"
		local result = {}
		local attrs
		local line
		local pathname
	
		local allEntries = {}
		local entry
		for entry in lfs.dir( path ) do
			if entry ~= "." and entry ~= ".." and (not filter or entry:match(filter)) then
				line = path .. "/" .. entry
				attrs = lfs.attributes(line)
				if attrs and attrs.mode == "directory" then
					table.insert( allEntries, {
						TIME = attrs.change,
						FULL = line,
						DIRNAME = entry
					})
				end
			end
		end
	
		table.sort( allEntries, function(a,b)
			return a.TIME < b.TIME
		end)

		local k,v
		for k,v in pairs( allEntries ) do
			line = v.FULL
			pathname = v.DIRNAME
	
			local stream = pathname:match( "([msMS]*)-[0-9_%-%:]+" )
			
			if stream then
				stream = stream:upper()
				local sequence = pathname
				local item
				local startTime = nil
				local endTime = nil
				local format = nil
				local snapfile = nil
				local file
	
				file = io.open( line .. "/.START", "r" )
				if file then
					startTime = tonumber( file:read("l") ) or 0
					file:close()
				end
	
				file = io.open( line .. "/.END", "r" )
				if file then
					endTime = tonumber( file:read("l") ) or 0
					file:close()
				end
	
				file = io.open( line .. "/.FORMAT", "r" )
				if file then
					format = file:read("l")
					file:close()
				end
	
				file = io.open( line .. "/SNAPSHOT.jpg", "r" )
				if file then
					snapfile = line .. "/SNAPSHOT.jpg"
					file:close()
				end
	
				local filelist = {}
				local totalSize = 0
				local fileCounts = 0
	
				for file in lfs.dir( line ) do
					local tmpf = file:upper()
	
					if tmpf ~= "." and tmpf ~= ".." and tmpf ~= ".START" and tmpf ~= ".END" and tmpf ~= ".FORMAT" and tmpf ~= "SNAPSHOT.JPG" and not tmpf:match("^%.") then
						attrs = lfs.attributes(line .. "/" .. file)
						if attrs and attrs.mode == "file" then 
							if not format then
								local filext = file:match( ".+%.(.*)" ) or ""
								format = "unknown"
								filext = filext:lower()
								if filext == "ts" then
									format = "ts"
								elseif filext == "mp4" then
									format = "mp4"
								elseif filext == "mkv" then
									format = "mkv"
								elseif filext == "mov" then
									format = "mov"
								elseif filext == "avi" then
									format = "avi"
								elseif filext == "264" or filext == "h264" then
									format = "264"
								end
							end
		
							--local size = (attrs.size + 1023)/1024  --Get the size in KB.
							--BUG fixed: Because the LUA large number problem, I need get file size by blocks:
							local size = 0
							if attrs.blocks == 0 then
								-- BUG fixed: Recording file in CIFS filesystem get the blocks is always 0.
								size = (attrs.size + 1023)/1024
							else
								size = (attrs.blocks*512.0 + 1023)/1024
							end
							-- local size = (attrs.blocks*512.0 + 1023)/1024 --Get the size in KB.
							filelist[file] = size
							fileCounts = fileCounts + 1
							totalSize = totalSize + size
						end
					end
				end
	
				local rec = {
					DEVICE = device,
					PATH = line,
					STREAM = (stream == "M" and "main" or "sub"), --NOTE: only consider main and sub.
					SEQUENCE = sequence,
					FILES = "",
					FILE_COUNT = fileCounts,
					START_TIME = startTime,
					END_TIME = endTime,
					FORMAT = format,
					SIZE = totalSize,
					SNAPSHOT = snapfile
				}
	
				rec.FILELIST = {}
				for kkk,vvv in pairs(filelist) do
					rec.FILES = rec.FILES .. kkk .. "=" .. (vvv or 0) .. "\n"
					table.insert(rec.FILELIST, {name = kkk, size = vvv})
				end
	
				local recordItem = _M.RECORDING[ stream == "M" and 0 or 1 ]
				if (recordItem.Status == "started" or recordItem.Status == "starting") and recordItem.SequenceID == sequence then
					rec._CURRENT = true
				end
	
				table.insert( result, rec )
			end --If get stream id and sequence.
		end --For each entry.
		device_result[device] = result
	end

	if type(device) ~= "table" then
		return device_result[device]
	end
	return device_result
end
----------------------------------------------------
local removingFiles = {}

function  _M.Isintable(value,tb)
    for k,v in pairs(tb) do
        if v == value then
            return true
        end
    end
    return false    --重点：全部跑完以后，如果非true，则返回false
end

function _M.ReserveDiskSpace( streamIdx, sizeMB )
	if streamIdx < 0 or streamIdx > 1 then
		print("return because streamIdx invalid:", streamIdx)
		return nil
	end
	
	local recordItem = _M.RECORDING[streamIdx]
	if not recordItem or not recordItem.NasDevice or not next(recordItem.NasDevice) then
		return
	end

	local fix = {}
	for i = #recordItem.NasDevice, 1, -1 do
		local free = _M.ALL_DISKS[recordItem.NasDevice[i]].FREE or 0
		if free < sizeMB then
			table.insert(fix, {device = recordItem.NasDevice[i], idx = i, free = free})
		end
	end

	if #fix == 0 then
		return true
	end

	if  _G.record_m[streamIdx].disk_policy ~= "overwrite" then
		return
	end

	for _, v in ipairs(fix) do
		local rec_list = _M.CheckRecordList( v.device )
		--First, try drop current stream records.
		local skip
		local stream = "main"
		if streamIdx > 0 then
			stream = "sub"
		end

		local file_name_t = {}
		for k,v in pairs(_M.RECORDING) do
			table.insert(file_name_t, v.CURRENT_FILE)
		end

		for i=1, #rec_list do
			for _,file in ipairs(rec_list[i].FILELIST) do
				if not rec_list[i]._CURRENT then	--Not current file.
					--if rec_list[i].STREAM == stream then	--Matched.
						print( "*** REMOVE(Overwrite) old records: " .. rec_list[i].PATH )
						local task = run_daemon.Create({
							name = "Delete old file: " .. rec_list[i].PATH ,
							run = "/bin/rm",
							args = {"-Rf", rec_list[i].PATH},
						})
						if task then
							task:Start()
						end
						rec_list[i]._KICKED = true
						v.free = v.free + (rec_list[i].SIZE/1024)	--To MB(size is KB)
					--end
				else
					path = rec_list[i].PATH .. "/" .. file.name
					if _M.Isintable(file.name,file_name_t) == false then
						print( "*** REMOVE(Overwrite) old2 records: " .. path )
						local task = run_daemon.Create({
							name = "Delete old file: " .. path ,
							run = "/bin/rm",
							args = {"-Rf", path},
							on_terminate = function()
								removingFiles[v.device][file.name] = nil
							end,
						})
						if task then
							task:Start()
							removingFiles[v.device] = removingFiles[v.device] or {}
							removingFiles[v.device][file.name] = true
						end
						rec_list[i]._KICKED = true
						v.free = v.free + (rec_list[i].SIZE/1024)	--To MB(size is KB)
					end
				end
			end

			if v.free >= sizeMB then
				skip = true
				break
			end
		end

		-- if not skip then
		-- 	--Then, try drop other streams record.
		-- 	for i=1, #rec_list do
		-- 		if not rec_list[i]._CURRENT and not rec_list[i]._KICKED then	--Not current file, and not processed.
		-- 			print( "*** REMOVE(Overwrite) old records: " .. rec_list[i].PATH )
		-- 			local r,s,c = os.execute( "/bin/rm -Rf " .. rec_list[i].PATH )
		-- 			if r and s == "exit" and c == 0 then
		-- 				rec_list[i]._KICKED = true
		-- 				v.free = v.free + (rec_list[i].SIZE/1024)	--To MB(size is KB)
		-- 			end
		-- 		end
	
		-- 		if v.free >= sizeMB then
		-- 			skip = true
		-- 			break
		-- 		end
		-- 	end
		-- end

		if not skip then
			print("OCLLOG, not skip and return nil from CheckDiskSize")
			-- table.remove(recordItem.NasDevice, v.idx)
			return nil
		end
	end
	return #recordItem.NasDevice ~= 0
end

----------------------------------------------------
-- local removingFiles = {}
local function FormatTime(timestamp)
	if not tonumber(timestamp) then
		return ""
	end
	local date = os.date("*t", timestamp)
	return string.format("%04d-%02d-%02d %02d:%02d:%02d", date.year, date.month, date.day, date.hour, date.min, date.sec)
end
function _M.GetDiskFiles(device)
	local result = _M.CheckRecordList(device)
	local files = {}
	for _, v in ipairs(result) do
		for _, file in ipairs(v.FILELIST) do
			if not removingFiles[device] or not removingFiles[device][file.name] then
				table.insert(files, {
					stream = v.STREAM,
					startTime = FormatTime(v.START_TIME),
					endTime = FormatTime(v.END_TIME),
					-- size = v.SIZE,
					size = file.size,
					name = file.name,
					path = v.PATH .. "/" .. file.name,
					current = v._CURRENT or false
				})
			end
		end
	end
	return files
end

local TOOLS = require "system.cmdtools"
local function EnumDisks()
	local rslt = {}
	local name
	for name in lfs.dir( "/sys/class/block" ) do
		if name ~= "." and name ~= ".." then
			if name:match("loop") or name:match("mtd") or name:match("ram") or name:match("rom") then
				--_dbg.Info( "SYSCTRL", "Enum skip block device of", name )
			else
				local pline
				local pfile = io.popen( TOOLS.UDEVADM .. " info --query=all --name=/dev/" .. name, "r" )
				if pfile then
					local line = pfile:read("*l")
					while line do
						pline = line:match("^P:%s+(.+)$")
						if pline then
							break
						end
						line = pfile:read("*l")
					end
					pfile:close()
				end

				if pline then
					if not _M.ALL_DISKS[name] then
						_M.DiskAttach(name, "usb", "")
					end
				end
			end
			--Found block device.
		end
		--Valid name
	end --for.
end
----------------------------------------------------
-- Step check to update recording [ Generally, I need to reserve disk spaces ]
-- The master control program need call this period, ~about 5 seconds each time.

function _M.StepCheck()
	local k,v
	--Update the disk free spaces first.
	for k,v in pairs(_M.ALL_DISKS) do
		local done, total, used, free, percent = _M.GetDiskSpaces( k, v.TYPE == "nas" )
		if done then
			--print( "Disk '" .. device .. "' info: TOTAL=" .. total .. " MB, USED=" .. used .. " MB, FREE=" .. free .. " MB, Percent=" .. percent .. "%" ) 
			v.TOTAL = total
			v.USED = used
			v.FREE = free
			v.PERCENT = percent
		end
	end

	for k,v in pairs(_M.RECORDING) do
	-- local k, v = next(_M.RECORDING)
		if v.Status == "started" then
			local limit_type = _G.record_m[k].limit_type
			
			if v.action == nil then
				v.action_timeout = v.action_timeout and (v.action_timeout + 1) or 0
			end

			if v.action_timeout == 4 or (v.action == 'end' and (limit_type == 'size' or limit_type == 'time')) then
				_M.StopRecord( k == 0 and "main" or "sub" )
				v.action_timeout = nil
			end

			if not _M.ReserveDiskSpace( k, 300 ) then
				print( "WARNING: Insufficent disk spaces for stream " .. k .. " recording, so stop the recording!" )
				v.ErrorAction = "recording"
				v.Error = "nospace"
				_M.StopRecord( k == 0 and "main" or "sub" )
			end
		end
	end

	EnumDisks()
end

function _M.RemoveFile(path)
	local device
	local tpe, f = path:match("%/media%/(.-)%/(.-)%/(.+)")
	if tpe == "nas" then
		device = "/media/" .. tpe .. "/" .. f
	else
		device = f
	end

	local result = _M.CheckRecordList(device)
	for _, v in ipairs(result) do
		for _, file in ipairs(v.FILELIST) do
			if path == v.PATH .. "/" .. file.name then
				if v._CURRENT then
					return false, "Can't Delete Recoding File"
				else
					local task = run_daemon.Create({
						name = "deleteFile",
						run = "/bin/rm",
						args = {path},
						on_terminate = function()
							removingFiles[device][file.name] = nil
						end,
					})
					if task then
						task:Start()
						removingFiles[device] = removingFiles[device] or {}
						removingFiles[device][file.name] = true
					else
						print("*** Error!!! run daemon failed!!!")
					end
					return true
				end
			end
		end
	end
	return true
end

local codec_stopd = {}
function _M.ControlRecord(stream, enable)
	if not enable then
		local _, doStop = _M.StopRecord(stream)
		codec_stopd[stream] = doStop
		return
	end

	if codec_stopd[stream] then
		codec_stopd[stream] = nil
		_M.StartRecord(stream)
	end
end

_M.Rule = {
	GetInfo = {
        is_open = true,
        parameters = {}
	},
	Update = {
        is_open = true,
        parameters = {
            {key = "cfg", type = "table", must = true},
        }
	},
	DiskList = {
        is_open = true,
        parameters = {}
	},
	GetDiskFiles = {
        is_open = true,
        parameters = {
            {key = "device", type = "string", must = true},
        }
	},
	ReserveDiskSpace = {
		is_open = true,
		parameters = {{key = "streamIdx", type = "number"}, {key = "sizeMB", type = "number"}}
	},
	CheckRecordList = {
		is_open = true,
		parameters = {{key = "device", type = "string"}, {key = "filter", type = "string"}}
	},
	InitRecordDB = {
		is_open = true,
		parameters = {{key = "device", type = "string"}}
	},
	StopRecord = {
		is_open = true,
		parameters = {{key = "stream", type = "string"}}
	},
	StartRecord = {
		is_open = true,
		parameters = {{key = "stream", type = "string"}}
	},
	PickDisk = {
		is_open = true,
		parameters = {{key = "req", type = "string"}}
	},
	GetDiskSpaces = {
		is_open = true,
		parameters = {{key = "device", type = "string"}, {key = "isNas", type = "boolean"}}
	},
	UmountDisk = {
		is_open = true,
		parameters = {{key = "device", type = "string"}}
	},
	MountDisk = {
		is_open = true,
		parameters = {{key = "device", type = "string"}, {key = "type", type = "string"}, {key = "filesys", type = "string"}}
	},
	GetStatus = {
		is_open = true,
		parameters = {{key = "stream", type = "string"}}
	},
	UpdateRecord = {
		is_open = true,
		parameters = {{key = "sequence_id", type = "string"}, {key = "action", type = "string"}, {key = "filename", type = "string"}, {key = "format", type = "string"},
		{key = "time", type = "string"}, {key = "size_kb", type = "number"},}
	},
	AckStopRecord = {
		is_open = true,
		parameters = {{key = "sequence_id", type = "string"}, {key = "ackCode", type = "string"}}
	},
	AckStartRecord = {
		is_open = true,
		parameters = {{key = "sequence_id", type = "string"}, {key = "ackCode", type = "string"}}
	},
	RegisterWorker = {
		is_open = true,
		parameters = {{key = "stream", type = "string"}}
	},
	DiskRecordList = {
		is_open = true,
		parameters = {{key = "device", type = "string"}, {key = "filter", type = "string"}}
	},
	DiskAttach = {
		is_open = true,
		parameters = {{key = "device", type = "string"}, {key = "type", type = "string"}, {key = "volName", type = "string"}, {key = "filesys", type = "string"}}
	},
	DiskDettach = {
		is_open = true,
		parameters = {{key = "device", type = "string"}}
	},
	RemoveFile = {
		is_open = true,
		parameters = {{key = "path", type = "string", must = true}}
	},
	ControlRecord = {
		is_open = true,
		parameters = {{key = "stream", type = "string", true}, {key = "enable", type = "boolean", true}}
	}

}

----------------------------------------------------
return _M
